package rs;
#
#Perl implementation of Reed Solomon Encoding for Postal-4i symbology
#
#Copyright (C) 2006 J. Wells
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without modification,
#are permitted provided that the following conditions are met:
#1. Redistributions of source code shall retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#2. Redistributions in binary form shall reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#3. The name of the author may not be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
#THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
#IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
#OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
#IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
#INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
#THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#This implementation is intended to provide an easy to follow illustration of the
#basic Reed-Solomon algorithm. It is designed for transparency, not efficiency!
#
#For reasons of transparency, the implementation is specific to the particular (63,51)
#case defined in the Postal-4i specification. For a more general C-language 
#implementation, the reader is referred to Phil Karn's implementation, downloadable
#from www.ka9q.net/code/fec/, which can be used under the terms of the GNU General 
#Public License (GPL). The version current at the time of development of this 
#implementation was reed-solomon-4.0.tar.gz
#
#usage:
	#create an array @c with c[0] to c[11] being 0; c[12] to c[63] the data to be encoded
	#execute @r = rs::encode(\@c)
	#returns resulting rs data in r[0] to r[11]

use Galoismath;		#use galois field arithmetic (assumes package rs is in Galoismath.pm)

sub rsinit {		#initialise rs algorithm parameters and tables
				#this subroutine is called once the first time rsencode is used
				#it sets up basic Postal-4i parameters and creates the
				#Galois Field parameters needed by associated package gf

	#basic reed solomon parameters
	$cwl = 6;				#no of bits in a codeword
	$gf::field = 63;		#GFno  = no of symbols in a message = 2**$cwl - 1
	$prim = 67;			#primitive polynomial = x^6 + x + 1

	$alpha = 2;			#base used for generating polynomial
	$initpwr = 1;			#power of alpha of first root of generating polynomial
	$nroots = 12;			#number of roots = no of ECC codewords

	#Generate Galois field lookup tables

	$gf::log[0] = $gf::field;		#log(0) = -inf
	$gf::alog[$gf::field] = 0;		#alpha**-inf = 0

	$sr = 1;
	for ($i=0; $i<$gf::field; $i++) {
		$gf::log[$sr] = $i;
		$gf::alog[$i] = $sr;
		$sr <<= 1;
		if ($sr & (1<<$cwl)) {$sr ^= $prim}
		$sr &= $gf::field;
	}


	#form RS code generator polynomial from its roots
	@g = (0) x ($nroots+1);	#create polynomial coefficients - 1 more than number of roots
	$g[0] = 1;				#initialise the first one to 1
	$root = $gf::alog[($initpwr * $gf::log[$alpha]) % $gf::field];	# 1st is $alpha**$fcr
	for ($i=0; $i<$nroots; $i++) {							#$nroot iterations needed
		#first do coefficients $nroots down to 1
		for ($j=$nroots; $j>0; $j--) {$g[$j] = gf::sum($g[$j-1],gf::prod($g[$j],$root))}
		$g[0] = gf::prod($g[0],$root);					#end with coefficient of x**0

		$root = gf::prod($root,$alpha);			#mutiply by $alpha to generate next root
	}

	#>>>>edit out hashes below to print out log, antilog tables and generating polynomial
	#print "log table is\t@gf::log\n";
	#print "antilog table is\t@gf::alog\n";
	#print "generating polynomial coefficients are\t@g\n";
}

sub encode {							#rs encoder

	my $dref = shift;					#pick up passed REFERENCE to coefficients
	my @d = @$dref;						#copy the array to avoid overwriting it

	unless (defined $nroots) {rsinit()}		#do initialisation if not already done

	for (my $i=$gf::field-1; $i>=$nroots; $i--) {		#from 62 down to 12
		for (my $j=0; $j<=$nroots; $j++) {			#divide by generating polynomial
			$d[$i+$j-$nroots] = gf::diff($d[$i+$j-$nroots],gf::prod($d[$i],$g[$j]));
		}
	}

	my @r = ();								#copy out the remainder
	for (my $i=0; $i<$nroots; $i++) {$r[$i] = $d[$i]}
	return @r;									#and return the result
}

1;	#so use succeeds if the package is put in a separate .pm file
